.g-frontpage-header
  padding 5px

  .g-frontpage-title-container
    display inline-block
    margin-left 12px

    .g-frontpage-title
      font-weight bold
      font-size 34px
      color #222

    .g-frontpage-subtitle
      font-size 18px
      color #707070
      margin-left 2px
      border-top 1px solid #eee

  .g-frontpage-logo-container
    position relative
    display inline-block
    vertical-align top
    width 82px
    height 82px

    .g-frontpage-logo
      max-width 82px
      max-height 82px
      // absolute allows top / transform to work
      position absolute
      // center image vertically and horizontally
      left 50%
      top 50%
      transform translate(-50%, -50%)

.g-frontpage-body
  padding 12px 12px 0
  margin-top 10px

  border-radius 2px
  background-color #f8f8f8
  border 1px solid #e5e5e5
  color #555

  .g-frontpage-paragraph
    margin-top 0
    margin-bottom 18px
    font-size 14px
