.g-current-assetstores-container
  padding-bottom 10px
  border-bottom 1px solid #d7d7d7

.g-assetstore-buttons
  margin-top 10px
  .btn
    margin-right 12px

#g-new-assetstore-container
  margin-top 10px

.g-assetstore-capacity-chart
  height 150px
  width 280px

.g-assetstore-container .panel-body
  &:before
    display none
  &:after
    display none

.g-assetstore-current-text
  font-weight bold
  margin-left 10px

.g-assetstore-current
  background rgba(50, 118, 177, 0.25)
  border-color #444

.g-assetstore-button-container
  display inline-block

.jqplot-data-label
  color black

.g-assetstore-import-button-container
  display inline-block

.g-import-instructions
  margin-bottom 22px
  padding-top 8px
  border-top 1px solid #eee

.g-submit-assetstore-import
  margin-top 7px
