.g-search-results-title-container
  display block
  padding 0.5% 0 0.5% 0.5%
  font-size 20px
  background-color #eaebea
  border-bottom 1px solid #ddd
  color #555

  .g-search-results-title-element
    display inline-block

.g-search-results-type-container
  display block
  margin 2px 0
  border 1px solid #efefef
  border-radius 5px

  .g-search-result-type-header
    display inline-block
    font-size 17px

  .g-search-result-type-icon
    padding 0.5% 1%

  ul.g-search-results-type
    margin 0
    padding 0

  .g-search-result
    list-style none
    padding 0.15% 0 0.15% 2%
    border-top 1px solid #efefef
    color #337ab7
    font-size 15px

    &:hover
      background-color #eee

  .g-search-results-paginate

    ul.pagination
      margin 5px 15px

.g-search-pending
  padding 10px
  font-size 30px

.g-search-no-results
  padding 0.5% 0 0.5% 1%
  border 1px solid #efefef
  font-size 15px
